#pragma once
#ifndef CommonShaders_h
#define CommonShaders_h

// This macro is used to define the shader codes as multi-line strings without quotation marks
// to allow for basic syntax highlighting in Visual Studio.
#define MULTILINE(...) #__VA_ARGS__

const char* CameraUniforms = MULTILINE(
	uniform mat4 viewMatrix;
	uniform mat4 projectionMatrix;
	uniform mat4 viewProjectionMatrix;
	uniform mat4 inverseViewMatrix;
	uniform vec2 viewPortSize;
	uniform vec4 cameraPosition;

	uniform vec4 frustumPlaneLeft;
	uniform vec4 frustumPlaneRight;
	uniform vec4 frustumPlaneBottom;
	uniform vec4 frustumPlaneTop;
	uniform vec4 frustumPlaneFar;
	uniform vec4 frustumPlaneNear;
);

const std::vector<const char*> NoiseCompute
{
	MULTILINE(
		#version 450\n
		#extension GL_NV_gpu_shader5 : enable\n

		layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

		layout(std430) restrict writeonly buffer heightfieldBufferBlock
		{
			float heightfieldBuffer[];
		};

		uniform float heightfieldScaleFactor;
		uniform vec2 cellSize;
		uniform ivec2 gridDimensions;
		uniform vec2 domainOrigin;
		uniform int numThreads;

		float rand(vec3 p3)
		{
			p3 = fract(p3 * 0.1031);
			p3 += dot(p3, p3.yzx + 19.19);

			return fract(dot(p3.xy, p3.zz));
		}

		float noiseIteration(vec3 x)
		{
			vec3 integral = floor(x);
			vec3 fractional = fract(x);
			vec3 weights = fractional * fractional * (-2.0 * fractional + 3.0);

			vec4 c = mix(
				vec4(
					rand(integral + vec3(0.0, 0.0, 0.0)),
					rand(integral + vec3(0.0, 1.0, 0.0)),
					rand(integral + vec3(0.0, 0.0, 1.0)),
					rand(integral + vec3(0.0, 1.0, 1.0))),
				vec4(
					rand(integral + vec3(1.0, 0.0, 0.0)),
					rand(integral + vec3(1.0, 1.0, 0.0)),
					rand(integral + vec3(1.0, 0.0, 1.0)),
					rand(integral + vec3(1.0, 1.0, 1.0))),
				weights.x);

			vec2 d = mix(c.xz, c.yw, weights.y);

			return mix(d.x, d.y, weights.z);
		}

		const int NOISE_ITERATIONS = 10;

		float noise(vec2 worldSpacePosition)
		{
			vec3 x = vec3(0.001 * worldSpacePosition, 0.0) + 13.37;
			float v = 0.0;
			float a = float(1 << (NOISE_ITERATIONS - 1));

			for (int i = 0; i < NOISE_ITERATIONS; i++)
			{
				v += a * noiseIteration(x);
				x *= 2.0;
				a *= 0.5;
			}

			return v / float((1 << NOISE_ITERATIONS) - 1);
		}

		void main()
		{
			int cellIndex = int(gl_GlobalInvocationID.x);

			if (cellIndex >= numThreads)
			{
				return;
			}

			int y = cellIndex / gridDimensions.x;
			int x = cellIndex - y * gridDimensions.x;

			vec2 cellCenter = domainOrigin + vec2(x, y) + 0.5 * cellSize;

			float heightfieldValue = heightfieldScaleFactor * noise(cellCenter);

			heightfieldBuffer[cellIndex] = heightfieldValue;
		}
	)
};

const char* HeightfieldSampling = MULTILINE(
	layout(std430) restrict readonly buffer heightfieldBufferBlock
	{
		float heightfieldBuffer[];
	};

	uniform float heightfieldScaleFactor;
	uniform ivec2 gridDimensions;
	uniform vec2 cellSize;
	uniform vec2 domainOrigin;

	float fetchHeightfieldValue(ivec2 gridIndex)
	{
		return heightfieldBuffer[gridIndex.y * gridDimensions.x + gridIndex.x];
	}

	float sampleHeightfieldNearestNeighbor(vec2 worldSpacePosition)
	{
		ivec2 gridIndex = clamp(ivec2((worldSpacePosition - domainOrigin) / cellSize), ivec2(0), gridDimensions - 1);

		return fetchHeightfieldValue(gridIndex);
	}

	float sampleHeightfieldBilinear(vec2 worldSpacePosition)
	{
		vec2 relativeSamplePosition = (worldSpacePosition - domainOrigin) / cellSize;
		ivec2 gridIndex = ivec2(relativeSamplePosition);

		vec2 cellCenter = (vec2(gridIndex) + 0.5) * cellSize;

		ivec2 bottomLeftOffset = mix(ivec2(0), ivec2(-1), lessThanEqual(worldSpacePosition, cellCenter));
		ivec2 bottomLeft = clamp(gridIndex + bottomLeftOffset, ivec2(0), gridDimensions - 1);
		ivec2 bottomRight = clamp(gridIndex + bottomLeftOffset + ivec2(1, 0), ivec2(0), gridDimensions - 1);
		ivec2 topLeft = clamp(gridIndex + bottomLeftOffset + ivec2(0, 1), ivec2(0), gridDimensions - 1);
		ivec2 topRight = clamp(gridIndex + bottomLeftOffset + ivec2(1, 1), ivec2(0), gridDimensions - 1);

		vec2 relativeCellCenter = vec2(bottomLeft) + 0.5;
		vec2 relativePositionInPatch = clamp(relativeSamplePosition - relativeCellCenter, 0.0, 1.0);

		float bottomLeftValue = fetchHeightfieldValue(bottomLeft);
		float bottomRightValue = fetchHeightfieldValue(bottomRight);
		float topLeftValue = fetchHeightfieldValue(topLeft);
		float topRightValue = fetchHeightfieldValue(topRight);

		return mix(
			mix(bottomLeftValue, bottomRightValue, relativePositionInPatch.x),
			mix(topLeftValue, topRightValue, relativePositionInPatch.x),
			relativePositionInPatch.y);
	}

	float sampleHeightfieldBicubic(vec2 worldSpacePosition)
	{
		vec2 relativeSamplePosition = (worldSpacePosition - domainOrigin) / cellSize;
		ivec2 gridIndex = ivec2(relativeSamplePosition);

		vec2 cellCenter = (vec2(gridIndex) + 0.5) * cellSize;

		ivec2 bottomLeftOffset = mix(ivec2(0), ivec2(-1), lessThanEqual(worldSpacePosition, cellCenter));
		ivec2 bottomLeft = clamp(gridIndex + bottomLeftOffset, ivec2(0), gridDimensions - 1);

		float values[16];

		for (int yOffset = -1; yOffset <= 2; yOffset++)
		{
			for (int xOffset = -1; xOffset <= 2; xOffset++)
			{
				ivec2 currentGridIndex = clamp(gridIndex + bottomLeftOffset + ivec2(xOffset, yOffset), ivec2(0), gridDimensions - 1);

				values[4 * (yOffset + 1) + xOffset + 1] = fetchHeightfieldValue(currentGridIndex);
			}
		}

		vec2 relativeCellCenter = vec2(bottomLeft) + 0.5;
		vec2 f = clamp(relativeSamplePosition - relativeCellCenter, 0.0, 1.0);
		vec2 f2 = f * f;
		vec2 f3 = f2 * f;

		vec2 w0 = -0.5 * (f3 + f) + f2;
		vec2 w1 = 1.5 * f3 - 2.5 * f2 + 1.0;
		vec2 w2 = -1.5 * f3 + 2.0 * f2 + 0.5 * f;
		vec2 w3 = 0.5 * (f3 - f2);

		return w0.y * (w0.x * values[0] + w1.x * values[1] + w2.x * values[2] + w3.x * values[3]) +
			w1.y * (w0.x * values[4] + w1.x * values[5] + w2.x * values[6] + w3.x * values[7]) +
			w2.y * (w0.x * values[8] + w1.x * values[9] + w2.x * values[10] + w3.x * values[11]) +
			w3.y * (w0.x * values[12] + w1.x * values[13] + w2.x * values[14] + w3.x * values[15]);
	}

	float interpolateQuintic(float v0, float v1, float v2, float v3, float v4, float v5, float t)
	{
		float h0 = 1.0 / 120.0 * (1.0 + t) * t * (1.0 - t) * (2.0 - t) * (3.0 - t) * v0;
		float h1 = -1.0 / 24.0 * (2.0 + t) * t * (1.0 - t) * (2.0 - t) * (3.0 - t) * v1;
		float h2 = 1.0 / 12.0 * (2.0 + t) * (1.0 + t) * (1.0 - t) * (2.0 - t) * (3.0 - t) * v2;
		float h3 = 1.0 / 12.0 * (2.0 + t) * (1.0 + t) * t * (2.0 - t) * (3.0 - t) * v3;
		float h4 = -1.0 / 24.0 * (2.0 + t) * (1.0 + t) * t * (1.0 - t) * (3.0 - t) * v4;
		float h5 = 1.0 / 120.0 * (2.0 + t) * (1.0 + t) * t * (1.0 - t) * (2.0 - t) * v5;

		return h0 + h1 + h2 + h3 + h4 + h5;
	}

	float sampleHeightfieldBiquintic(vec2 worldSpacePosition)
	{
		vec2 relativeSamplePosition = (worldSpacePosition - domainOrigin) / cellSize;
		ivec2 gridIndex = ivec2(relativeSamplePosition);

		vec2 cellCenter = (vec2(gridIndex) + 0.5) * cellSize;

		ivec2 bottomLeftOffset = mix(ivec2(0), ivec2(-1), lessThanEqual(worldSpacePosition, cellCenter));
		ivec2 bottomLeft = clamp(gridIndex + bottomLeftOffset, ivec2(0), gridDimensions - 1);

		float values[36];

		for (int yOffset = -2; yOffset <= 3; yOffset++)
		{
			for (int xOffset = -2; xOffset <= 3; xOffset++)
			{
				ivec2 currentGridIndex = clamp(gridIndex + bottomLeftOffset + ivec2(xOffset, yOffset), ivec2(0), gridDimensions - 1);

				values[6 * (yOffset + 2) + xOffset + 2] = fetchHeightfieldValue(currentGridIndex);
			}
		}

		vec2 relativeCellCenter = vec2(bottomLeft) + 0.5;
		vec2 relativePositionInPatch = clamp(relativeSamplePosition - relativeCellCenter, 0.0, 1.0);

		float y0 = interpolateQuintic(values[0], values[1], values[2], values[3], values[4], values[5], relativePositionInPatch.x);
		float y1 = interpolateQuintic(values[6], values[7], values[8], values[9], values[10], values[11], relativePositionInPatch.x);
		float y2 = interpolateQuintic(values[12], values[13], values[14], values[15], values[16], values[17], relativePositionInPatch.x);
		float y3 = interpolateQuintic(values[18], values[19], values[20], values[21], values[22], values[23], relativePositionInPatch.x);
		float y4 = interpolateQuintic(values[24], values[25], values[26], values[27], values[28], values[29], relativePositionInPatch.x);
		float y5 = interpolateQuintic(values[30], values[31], values[32], values[33], values[34], values[35], relativePositionInPatch.x);

		return interpolateQuintic(y0, y1, y2, y3, y4, y5, relativePositionInPatch.y);
	}

	float sampleHeightfield(vec2 worldSpacePosition)
	{
		return sampleHeightfieldBicubic(worldSpacePosition);
	}
);

const std::vector<const char*> CommonGeometryShader
{
	MULTILINE(
		#version 450\n

		layout(triangles) in;
		layout(triangle_strip, max_vertices = 3) out;

		out vec3 gsHeights;

		void main()
		{
			const vec3[3] heights = vec3[3](vec3(1.0, 0.0, 0.0), vec3(0.0, 1.0, 0.0), vec3(0.0, 0.0, 1.0));

			for (int i = 0; i < 3; i++)
			{
				gsHeights = heights[i];
				gl_Position = gl_in[i].gl_Position;

				EmitVertex();
			}

			EndPrimitive();
		}
	)
};

const std::vector<const char*> CommonFragmentShader
{
	MULTILINE(
		#version 450\n

		in vec3 gsHeights;

		out vec4 outColor;
	),
	HeightfieldSampling,
	MULTILINE(
		void main()
		{
			outColor = vec4(1.0);

			// Wireframe rendering, see http://developer.download.nvidia.com/whitepapers/2007/SDK10/SolidWireframe.pdf for an explanation

			vec3 ddxHeights = dFdx(gsHeights);
			vec3 ddyHeights = dFdy(gsHeights);
			vec3 eDists = gsHeights / sqrt(ddxHeights * ddxHeights + ddyHeights * ddyHeights);

			float dist = min(eDists.x, min(eDists.y, eDists.z));

			if (dist <= 1.5)
			{
				dist = clamp(dist + 0.5, 0.0, 2.0);

				outColor = mix(outColor, vec4(0.0, 0.0, 0.0, 1.0), exp2(-2.0 * dist * dist));
			}
		}
	)
};

#endif