#include "TessellationCornelEtAl2022.h"

#include <chrono>
#include <iostream>
#include <vector>

#include <GL/glew.h>

#include "../Heightfield.h"
#include "ShaderCodesCornelEtAl2022.h"

#include "../PipelineStatisticsQuery.h"

using namespace Eigen;

const int maximumNumberOfTriangles = 1 << 23;
const float averageEdgeLengthInPixels = 10.0f;
const float minimumEdgeLengthInMeters = 1.0f;
const long long timeBudgetInMicroseconds = 5000ll;

const int FLAG_SUBDIVIDE = 1 << 0;
const int FLAG_MERGE = 1 << 1;
const int FLAG_DELETED = 1 << 2;

//=================================================================================================
/**
*	Constructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationCornelEtAl2022::TessellationCornelEtAl2022(const Heightfield& heightfield) : AbstractTessellation(heightfield, "CornelEtAl2022")
{
	//=================================================================================================
	/**
	*	Create the per-triangle buffers, the vertex buffer, and free indices buffers
	*/
	//=================================================================================================

	glCreateBuffers(1, &mSubdivisionFlagsBuffer);
	glNamedBufferData(mSubdivisionFlagsBuffer, sizeof(GLubyte) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	GLubyte deletedFlag = FLAG_DELETED;
	glClearNamedBufferData(mSubdivisionFlagsBuffer, GL_R8UI, GL_RED_INTEGER, GL_UNSIGNED_BYTE, &deletedFlag);

	glCreateBuffers(1, &mParentTriangleBuffer);
	glNamedBufferData(mParentTriangleBuffer, sizeof(GLint) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mFirstChildTriangleBuffer);
	glNamedBufferData(mFirstChildTriangleBuffer, sizeof(GLint) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mAdjacentTrianglesBuffer);
	glNamedBufferData(mAdjacentTrianglesBuffer, sizeof(Vector3i) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mVertexIndicesBuffer);
	glNamedBufferData(mVertexIndicesBuffer, sizeof(Vector3i) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mTriangleGraphColorBuffer);
	glNamedBufferData(mTriangleGraphColorBuffer, sizeof(GLubyte) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mVertexBuffer);
	glNamedBufferData(mVertexBuffer, sizeof(Vector3f) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mFreeTriangleIndicesBuffer);
	glNamedBufferData(mFreeTriangleIndicesBuffer, sizeof(GLint) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mFreeVertexIndicesBuffer);
	glNamedBufferData(mFreeVertexIndicesBuffer, sizeof(GLint) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	//=================================================================================================
	/**
	*	Create tessellation resources
	*/
	//=================================================================================================

	mUpdateSubdivisionFlagsProgram = std::make_unique<Program>(ShaderCodesCornelEtAl2022::UpdateSubdivisionFlagsCompute);
	mUpdateSubdivisionFlagsProgram->setUniform1f("squaredMinimumEdgeLength", minimumEdgeLengthInMeters * minimumEdgeLengthInMeters);
	mUpdateSubdivisionFlagsProgram->setUniform1f("upperTriangleSize", 1.5f * averageEdgeLengthInPixels);
	mUpdateSubdivisionFlagsProgram->setUniform1f("lowerTriangleSize", 0.75f * averageEdgeLengthInPixels);

	mSubdivisionTriangleSelectionProgram = std::make_unique<Program>(ShaderCodesCornelEtAl2022::SubdivisionTriangleSelectionCompute);

	mSubdivisionProgram = std::make_unique<Program>(ShaderCodesCornelEtAl2022::SubdivisionCompute);
	mSubdivisionProgram->setUniform2i("gridDimensions", mHeightfield.mDimensions);
	mSubdivisionProgram->setUniform2f("domainOrigin", mHeightfield.mOrigin);
	mSubdivisionProgram->setUniform2f("cellSize", mHeightfield.mCellSize);
	mSubdivisionProgram->setUniform1f("heightfieldScaleFactor", mHeightfield.mScaleFactor);
	mSubdivisionProgram->setUniform1f("squaredMinimumEdgeLength", minimumEdgeLengthInMeters * minimumEdgeLengthInMeters);
	mSubdivisionProgram->setUniform1f("upperTriangleSize", 1.5f * averageEdgeLengthInPixels);
	mSubdivisionProgram->setUniform1f("lowerTriangleSize", 0.75f * averageEdgeLengthInPixels);

	mMergeTriangleSelectionProgram = std::make_unique<Program>(ShaderCodesCornelEtAl2022::MergeTriangleSelectionCompute);

	mMergeProgram = std::make_unique<Program>(ShaderCodesCornelEtAl2022::MergeCompute);

	//=================================================================================================
	/**
	*	Create rendering resources
	*/
	//=================================================================================================

	mCompactionProgram = std::make_unique<Program>(ShaderCodesCornelEtAl2022::CompactionCompute);
	mRenderingProgram = std::make_unique<Program>(
		ShaderCodesCornelEtAl2022::RenderingVertex,
		ShaderCodesCornelEtAl2022::CommonGeometryShader,
		ShaderCodesCornelEtAl2022::CommonFragmentShader);

	for (int graphColorIteration = 0; graphColorIteration < 4; graphColorIteration++)
	{
		glCreateBuffers(1, &mGraphColorBuffers[graphColorIteration]);
		glNamedBufferData(mGraphColorBuffers[graphColorIteration], sizeof(int) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

		glCreateBuffers(1, &mGraphColorCounters[graphColorIteration]);
		glNamedBufferData(mGraphColorCounters[graphColorIteration], sizeof(GLint), nullptr, GL_DYNAMIC_COPY);
	}

	glCreateBuffers(1, &mIndirectDrawBuffer);
	glNamedBufferData(mIndirectDrawBuffer, sizeof(DrawElementsIndirectCommand), nullptr, GL_DYNAMIC_COPY);

	glCreateBuffers(1, &mVisibleTriangleBuffer);
	glNamedBufferData(mVisibleTriangleBuffer, sizeof(Vector3i) * maximumNumberOfTriangles, nullptr, GL_DYNAMIC_COPY);

	glCreateVertexArrays(1, &mVAO);
	glVertexArrayElementBuffer(mVAO, mVisibleTriangleBuffer);

	GLuint positionLocation = mRenderingProgram->getAttributeLocation("inPosition");
	glEnableVertexArrayAttrib(mVAO, positionLocation);
	glVertexArrayAttribFormat(mVAO, positionLocation, 3, GL_FLOAT, GL_FALSE, 0);
	glVertexArrayVertexBuffer(mVAO, positionLocation, mVertexBuffer, 0, sizeof(Vector3f));

	//=================================================================================================
	/**
	*	Create a coarse initial tessellation of the provided heightfield
	*/
	//=================================================================================================

	const ChunkDescriptor& chunkDescriptor = mHeightfield.mChunkDescriptor;

	int numTriangles = 2 * chunkDescriptor.mChunkGridDimensions.x() * chunkDescriptor.mChunkGridDimensions.y();
	int numVertices = (chunkDescriptor.mChunkGridDimensions.x() + 1) * (chunkDescriptor.mChunkGridDimensions.y() + 1);

	mUnusedTrianglesOffset = numTriangles;
	mNumReusableTriangleIndices = 0;
	mUnusedVerticesOffset = numVertices;
	mNumReusableVertexIndices = 0;

	Program initialTessellationProgram(ShaderCodesCornelEtAl2022::InitialTessellationCompute);
	initialTessellationProgram.bind();
	initialTessellationProgram.setSSBO("vertexBuffer", mVertexBuffer);
	initialTessellationProgram.setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	initialTessellationProgram.setSSBO("parentTriangleBuffer", mParentTriangleBuffer);
	initialTessellationProgram.setSSBO("firstChildTriangleBuffer", mFirstChildTriangleBuffer);
	initialTessellationProgram.setSSBO("triangleGraphColorBuffer", mTriangleGraphColorBuffer);
	initialTessellationProgram.setSSBO("adjacentTrianglesBuffer", mAdjacentTrianglesBuffer);
	initialTessellationProgram.setSSBO("vertexIndicesBuffer", mVertexIndicesBuffer);
	initialTessellationProgram.setSSBO("heightfieldBuffer", mHeightfield.mBuffer);
	initialTessellationProgram.setUniform1f("heightfieldScaleFactor", mHeightfield.mScaleFactor);
	initialTessellationProgram.setUniform2i("gridDimensions", mHeightfield.mDimensions);
	initialTessellationProgram.setUniform2f("domainOrigin", mHeightfield.mOrigin);
	initialTessellationProgram.setUniform2f("cellSize", mHeightfield.mCellSize);
	initialTessellationProgram.setUniform2f("domainUpper", mHeightfield.mOrigin + mHeightfield.mCellSize.cwiseProduct(mHeightfield.mDimensions.cast<float>()));
	initialTessellationProgram.setUniform2f("chunkGridCellSize", chunkDescriptor.mChunkCellSize);
	initialTessellationProgram.setUniform2i("chunkGridDimensions", chunkDescriptor.mChunkGridDimensions);
	initialTessellationProgram.compute1D(chunkDescriptor.mChunkGridDimensions.x() * chunkDescriptor.mChunkGridDimensions.y());
	initialTessellationProgram.unbind();

	glMemoryBarrier(GL_ALL_BARRIER_BITS);
}

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationCornelEtAl2022::~TessellationCornelEtAl2022()
{
	glDeleteBuffers(1, &mSubdivisionFlagsBuffer);
	glDeleteBuffers(1, &mParentTriangleBuffer);
	glDeleteBuffers(1, &mFirstChildTriangleBuffer);
	glDeleteBuffers(1, &mAdjacentTrianglesBuffer);
	glDeleteBuffers(1, &mVertexIndicesBuffer);
	glDeleteBuffers(1, &mTriangleGraphColorBuffer);
	glDeleteBuffers(1, &mVertexBuffer);
	glDeleteBuffers(1, &mFreeTriangleIndicesBuffer);
	glDeleteBuffers(1, &mFreeVertexIndicesBuffer);

	for (int graphColorIteration = 0; graphColorIteration < 4; graphColorIteration++)
	{
		glDeleteBuffers(1, &mGraphColorBuffers[graphColorIteration]);
		glDeleteBuffers(1, &mGraphColorCounters[graphColorIteration]);
	}

	glDeleteBuffers(1, &mIndirectDrawBuffer);
	glDeleteBuffers(1, &mVisibleTriangleBuffer);
	glDeleteVertexArrays(1, &mVAO);
}

//=================================================================================================
/**
*	Performs a per-frame update of the tessellation.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationCornelEtAl2022::update(const Camera& camera)
{
	if (mPreviousViewProjectionMatrix.isApprox(camera.mViewProjectionMatrix))
	{
		return;
	}
	else
	{
		mPreviousViewProjectionMatrix = camera.mViewProjectionMatrix;
	}

	//=================================================================================================
	/**
	*	Update of triangle subdivision flags
	*/
	//=================================================================================================

	updateSubdivisionFlags(camera);

	//=================================================================================================
	/**
	*	Tessellation loop
	*/
	//=================================================================================================

	bool hasChanges = false;
	bool hasAnyChanges = false;
	bool iterate = true;

	auto iterationTimer = std::chrono::system_clock::now();

	// This map is used to identify whether the same changes are requested over and over again
	std::unordered_map<int, int> changesTracker;

	do
	{
		int requiredResizeElements = 0;
		int numSelectedTrianglesSubdivision = 0;
		int numSelectedTrianglesMerging = 0;

		hasChanges = subdivideTriangles(camera, numSelectedTrianglesSubdivision, requiredResizeElements);

		bool resizeRequired = (requiredResizeElements > 0);

		if (resizeRequired)
		{
			// The triangle buffer is full. Depending on the application, resize the buffer here or ignore it.
			// No further subdivision will occur until triangles have been freed by merging.
		}
		else
		{
			hasChanges = (mergeTriangles(camera, numSelectedTrianglesMerging)) || (hasChanges);
		}

		// The tessellation iteration has completed, now we only check whether to continue or terminate the loop

		hasAnyChanges = (hasAnyChanges) || (hasChanges);

		int numChanges = std::max(numSelectedTrianglesSubdivision, numSelectedTrianglesMerging);
		bool significantChange = (mFirstFrame) || (numChanges >= 100);
		bool repetitiveChange = false;

		if ((hasChanges) && (!resizeRequired) && (mFirstFrame))
		{
			auto it = changesTracker.find(numChanges);

			if (it == changesTracker.end())
			{
				changesTracker[numChanges] = 1;
			}
			else
			{
				if (it->second >= 3)
				{
					repetitiveChange = true;
				}
				else
				{
					it->second++;
				}
			}
		}

		auto end = std::chrono::system_clock::now();
		long long iterationTime = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::system_clock::now() - iterationTimer).count();

		iterate = (hasChanges) && (!repetitiveChange) && (significantChange) && ((mFirstFrame) || (iterationTime <= timeBudgetInMicroseconds));
	} while (iterate);

	if ((mFirstFrame) || (hasAnyChanges))
	{
		compactTriangles(camera);
	}

	mFirstFrame = false;
}

//=================================================================================================
/**
*	Renders the geometry.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationCornelEtAl2022::render(const Camera& camera)
{
	glBindVertexArray(mVAO);

	glBindBuffer(GL_DRAW_INDIRECT_BUFFER, mIndirectDrawBuffer);

	mRenderingProgram->bind();
	mRenderingProgram->setCamera(camera);

	glMultiDrawElementsIndirect(GL_TRIANGLES, GL_UNSIGNED_INT, 0, 1, 0);

	mRenderingProgram->unbind();

	glBindBuffer(GL_DRAW_INDIRECT_BUFFER, 0u);

	glBindVertexArray(0u);
}

//=================================================================================================
/**
*	Updates the subdivision flags of all triangles based on their distance to the camera.
*
*	@author	Daniel Cornel
*/
//=================================================================================================
void TessellationCornelEtAl2022::updateSubdivisionFlags(const Camera& camera)
{
	mUpdateSubdivisionFlagsProgram->bind();
	mUpdateSubdivisionFlagsProgram->setCamera(camera);
	mUpdateSubdivisionFlagsProgram->setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	mUpdateSubdivisionFlagsProgram->setSSBO("vertexIndicesBuffer", mVertexIndicesBuffer);
	mUpdateSubdivisionFlagsProgram->setSSBO("vertexBuffer", mVertexBuffer);
	mUpdateSubdivisionFlagsProgram->compute1D(mUnusedTrianglesOffset);
	mUpdateSubdivisionFlagsProgram->unbind();

	glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
}

//=================================================================================================
/**
*	Subdivides all triangles flagged for subdivision. Outputs the number of triangles that could
*	not be created because the triangle buffer is full. Returns whether any triangles have been
*	subdivided.
*
*	@author	Daniel Cornel
*
*	@param[out]	outNumSelectedTriangles		The number of triangles selected for subdivision
*	@param[out]	outRequiredResizeElements	The number of triangles that could not be created
*											because the triangle buffers are full
*/
//=================================================================================================
bool TessellationCornelEtAl2022::subdivideTriangles(const Camera& camera, int& outNumSelectedTriangles, int& outRequiredResizeElements)
{
	for (int graphColorIteration = 0; graphColorIteration < 4; graphColorIteration++)
	{
		glClearNamedBufferData(mGraphColorCounters[graphColorIteration], GL_R32I, GL_RED_INTEGER, GL_INT, nullptr);
	}

	mSubdivisionTriangleSelectionProgram->bind();
	mSubdivisionTriangleSelectionProgram->setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	mSubdivisionTriangleSelectionProgram->setSSBO("firstChildTriangleBuffer", mFirstChildTriangleBuffer);
	mSubdivisionTriangleSelectionProgram->setSSBO("adjacentTrianglesBuffer", mAdjacentTrianglesBuffer);
	mSubdivisionTriangleSelectionProgram->setSSBO("triangleGraphColorBuffer", mTriangleGraphColorBuffer);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor0Buffer", mGraphColorBuffers[0]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor0Counter", mGraphColorCounters[0]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor1Buffer", mGraphColorBuffers[1]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor1Counter", mGraphColorCounters[1]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor2Buffer", mGraphColorBuffers[2]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor2Counter", mGraphColorCounters[2]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor3Buffer", mGraphColorBuffers[3]);
	mSubdivisionTriangleSelectionProgram->setSSBO("graphColor3Counter", mGraphColorCounters[3]);
	mSubdivisionTriangleSelectionProgram->compute1D(mUnusedTrianglesOffset);
	mSubdivisionTriangleSelectionProgram->unbind();

	glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT | GL_BUFFER_UPDATE_BARRIER_BIT);

	Vector4i selectedTrianglesPerGraphColor = Vector4i::Zero();

	for (int graphColorIteration = 0; graphColorIteration < 4; graphColorIteration++)
	{
		glGetNamedBufferSubData(mGraphColorCounters[graphColorIteration], 0, sizeof(GLint), &selectedTrianglesPerGraphColor[graphColorIteration]);
	}

	int numSelectedTriangles = selectedTrianglesPerGraphColor.sum();

	outNumSelectedTriangles = numSelectedTriangles;

	if (numSelectedTriangles <= 0)
	{
		outRequiredResizeElements = 0;

		return false;
	}

	int numReusableTriangleIndices = mNumReusableTriangleIndices;
	int unusedTrianglesOffset = mUnusedTrianglesOffset;
	unusedTrianglesOffset += 4 - unusedTrianglesOffset % 4;

	int availableReusableTriangleIndices = numReusableTriangleIndices;
	Vector4i numReusedTriangleIndicesPerGraphColor = Vector4i::Zero();
	Vector4i reusableTriangleIndicesOffsetPerGraphColor = Vector4i::Zero();

	reusableTriangleIndicesOffsetPerGraphColor[0] = std::max(0, availableReusableTriangleIndices - 1);
	numReusedTriangleIndicesPerGraphColor[0] = std::max(0, std::min(selectedTrianglesPerGraphColor[0], availableReusableTriangleIndices));
	availableReusableTriangleIndices -= numReusedTriangleIndicesPerGraphColor[0];

	reusableTriangleIndicesOffsetPerGraphColor[1] = std::max(0, availableReusableTriangleIndices - 1);
	numReusedTriangleIndicesPerGraphColor[1] = std::max(0, std::min(selectedTrianglesPerGraphColor[1], availableReusableTriangleIndices));
	availableReusableTriangleIndices -= numReusedTriangleIndicesPerGraphColor[1];

	reusableTriangleIndicesOffsetPerGraphColor[2] = std::max(0, availableReusableTriangleIndices - 1);
	numReusedTriangleIndicesPerGraphColor[2] = std::max(0, std::min(selectedTrianglesPerGraphColor[2], availableReusableTriangleIndices));
	availableReusableTriangleIndices -= numReusedTriangleIndicesPerGraphColor[2];

	reusableTriangleIndicesOffsetPerGraphColor[3] = std::max(0, availableReusableTriangleIndices - 1);
	numReusedTriangleIndicesPerGraphColor[3] = std::max(0, std::min(selectedTrianglesPerGraphColor[3], availableReusableTriangleIndices));

	int availableUnusedTriangles = std::max(0, (maximumNumberOfTriangles - unusedTrianglesOffset) / 4);
	int currentUnusedTrianglesOffset = unusedTrianglesOffset;
	Vector4i numUnusedTrianglesPerGraphColor = Vector4i::Zero();
	Vector4i unusedTrianglesOffsetPerGraphColor = Vector4i::Zero();

	numUnusedTrianglesPerGraphColor[0] = std::max(0, std::min(selectedTrianglesPerGraphColor[0] - numReusedTriangleIndicesPerGraphColor[0], availableUnusedTriangles));
	availableUnusedTriangles -= numUnusedTrianglesPerGraphColor[0];
	unusedTrianglesOffsetPerGraphColor[0] = currentUnusedTrianglesOffset;
	currentUnusedTrianglesOffset += 4 * numUnusedTrianglesPerGraphColor[0];

	numUnusedTrianglesPerGraphColor[1] = std::max(0, std::min(selectedTrianglesPerGraphColor[1] - numReusedTriangleIndicesPerGraphColor[1], availableUnusedTriangles));
	availableUnusedTriangles -= numUnusedTrianglesPerGraphColor[1];
	unusedTrianglesOffsetPerGraphColor[1] = currentUnusedTrianglesOffset;
	currentUnusedTrianglesOffset += 4 * numUnusedTrianglesPerGraphColor[1];

	numUnusedTrianglesPerGraphColor[2] = std::max(0, std::min(selectedTrianglesPerGraphColor[2] - numReusedTriangleIndicesPerGraphColor[2], availableUnusedTriangles));
	availableUnusedTriangles -= numUnusedTrianglesPerGraphColor[2];
	unusedTrianglesOffsetPerGraphColor[2] = currentUnusedTrianglesOffset;
	currentUnusedTrianglesOffset += 4 * numUnusedTrianglesPerGraphColor[2];

	numUnusedTrianglesPerGraphColor[3] = std::max(0, std::min(selectedTrianglesPerGraphColor[3] - numReusedTriangleIndicesPerGraphColor[3], availableUnusedTriangles));
	unusedTrianglesOffsetPerGraphColor[3] = currentUnusedTrianglesOffset;

	int triangleTotalUsedInterleaved = numReusedTriangleIndicesPerGraphColor.sum();
	int triangleNewFreeInterleaved = std::max(0, numReusableTriangleIndices - triangleTotalUsedInterleaved);
	int triangleTotalUsedEnd = numUnusedTrianglesPerGraphColor.sum();
	int triangleNewFreeEnd = unusedTrianglesOffset + 4 * triangleTotalUsedEnd;

	Vector4i graphColorNumThreads = numReusedTriangleIndicesPerGraphColor + numUnusedTrianglesPerGraphColor;

	outRequiredResizeElements = 4 * std::max(0, numSelectedTriangles - triangleTotalUsedInterleaved - triangleTotalUsedEnd);

	if (graphColorNumThreads.sum() <= 0)
	{
		return false;
	}

	int vertexEndOffset = mUnusedVerticesOffset;
	int vertexNumInterleaved = mNumReusableVertexIndices;

	int vertexAvailableFreeInterleaved = vertexNumInterleaved;
	Vector4i vertexGraphColorNumInterleaved = Vector4i::Zero();
	Vector4i vertexGraphColorInterleavedOffset = Vector4i::Zero();

	vertexGraphColorInterleavedOffset[0] = std::max(0, vertexAvailableFreeInterleaved - 1);
	vertexGraphColorNumInterleaved[0] = std::max(0, std::min(selectedTrianglesPerGraphColor[0], vertexAvailableFreeInterleaved));
	vertexAvailableFreeInterleaved -= vertexGraphColorNumInterleaved[0];

	vertexGraphColorInterleavedOffset[1] = std::max(0, vertexAvailableFreeInterleaved - 1);
	vertexGraphColorNumInterleaved[1] = std::max(0, std::min(selectedTrianglesPerGraphColor[1], vertexAvailableFreeInterleaved));
	vertexAvailableFreeInterleaved -= vertexGraphColorNumInterleaved[1];

	vertexGraphColorInterleavedOffset[2] = std::max(0, vertexAvailableFreeInterleaved - 1);
	vertexGraphColorNumInterleaved[2] = std::max(0, std::min(selectedTrianglesPerGraphColor[2], vertexAvailableFreeInterleaved));
	vertexAvailableFreeInterleaved -= vertexGraphColorNumInterleaved[2];

	vertexGraphColorInterleavedOffset[3] = std::max(0, vertexAvailableFreeInterleaved - 1);
	vertexGraphColorNumInterleaved[3] = std::max(0, std::min(selectedTrianglesPerGraphColor[3], vertexAvailableFreeInterleaved));

	int vertexAvailableFreeEnd = std::max(0, maximumNumberOfTriangles - vertexEndOffset);
	int vertexCurrentEndOffset = vertexEndOffset;
	Vector4i vertexGraphColorNumEnd = Vector4i::Zero();
	Vector4i vertexGraphColorEndOffset = Vector4i::Zero();

	vertexGraphColorNumEnd[0] = std::max(0, std::min(selectedTrianglesPerGraphColor[0] - vertexGraphColorNumInterleaved[0], vertexAvailableFreeEnd));
	vertexAvailableFreeEnd -= vertexGraphColorNumEnd[0];
	vertexGraphColorEndOffset[0] = vertexCurrentEndOffset;
	vertexCurrentEndOffset += vertexGraphColorNumEnd[0];

	vertexGraphColorNumEnd[1] = std::max(0, std::min(selectedTrianglesPerGraphColor[1] - vertexGraphColorNumInterleaved[1], vertexAvailableFreeEnd));
	vertexAvailableFreeEnd -= vertexGraphColorNumEnd[1];
	vertexGraphColorEndOffset[1] = vertexCurrentEndOffset;
	vertexCurrentEndOffset += vertexGraphColorNumEnd[1];

	vertexGraphColorNumEnd[2] = std::max(0, std::min(selectedTrianglesPerGraphColor[2] - vertexGraphColorNumInterleaved[2], vertexAvailableFreeEnd));
	vertexAvailableFreeEnd -= vertexGraphColorNumEnd[2];
	vertexGraphColorEndOffset[2] = vertexCurrentEndOffset;
	vertexCurrentEndOffset += vertexGraphColorNumEnd[2];

	vertexGraphColorNumEnd[3] = std::max(0, std::min(selectedTrianglesPerGraphColor[3] - vertexGraphColorNumInterleaved[3], vertexAvailableFreeEnd));
	vertexGraphColorEndOffset[3] = vertexCurrentEndOffset;

	int vertexTotalUsedInterleaved = vertexGraphColorNumInterleaved.sum();
	int vertexNewFreeInterleaved = std::max(0, vertexNumInterleaved - vertexTotalUsedInterleaved);
	int vertexTotalUsedEnd = vertexGraphColorNumEnd.sum();
	int vertexNewFreeEnd = vertexEndOffset + vertexTotalUsedEnd;

	mSubdivisionProgram->bind();
	mSubdivisionProgram->setCamera(camera);
	mSubdivisionProgram->setSSBO("heightfieldBuffer", mHeightfield.mBuffer);
	mSubdivisionProgram->setSSBO("vertexBuffer", mVertexBuffer);
	mSubdivisionProgram->setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	mSubdivisionProgram->setSSBO("firstChildTriangleBuffer", mFirstChildTriangleBuffer);
	mSubdivisionProgram->setSSBO("parentTriangleBuffer", mParentTriangleBuffer);
	mSubdivisionProgram->setSSBO("adjacentTrianglesBuffer", mAdjacentTrianglesBuffer);
	mSubdivisionProgram->setSSBO("triangleGraphColorBuffer", mTriangleGraphColorBuffer);
	mSubdivisionProgram->setSSBO("vertexIndicesBuffer", mVertexIndicesBuffer);
	mSubdivisionProgram->setSSBO("freeTriangleIndicesBuffer", mFreeTriangleIndicesBuffer);
	mSubdivisionProgram->setSSBO("freeVertexIndicesBuffer", mFreeVertexIndicesBuffer);

	for (int graphColor = 0; graphColor < 8; graphColor += 2)
	{
		int graphColorIteration = graphColor >> 1;
		int numThreads = graphColorNumThreads[graphColorIteration];

		if (numThreads > 0)
		{
			mSubdivisionProgram->setSSBO("graphColorBuffer", mGraphColorBuffers[graphColorIteration]);
			mSubdivisionProgram->setUniform1i("triangleGraphColorNumInterleaved", numReusedTriangleIndicesPerGraphColor[graphColorIteration]);
			mSubdivisionProgram->setUniform1i("triangleGraphColorInterleavedOffset", reusableTriangleIndicesOffsetPerGraphColor[graphColorIteration]);
			mSubdivisionProgram->setUniform1i("triangleGraphColorEndOffset", unusedTrianglesOffsetPerGraphColor[graphColorIteration]);
			mSubdivisionProgram->setUniform1i("vertexGraphColorNumInterleaved", vertexGraphColorNumInterleaved[graphColorIteration]);
			mSubdivisionProgram->setUniform1i("vertexGraphColorInterleavedOffset", vertexGraphColorInterleavedOffset[graphColorIteration]);
			mSubdivisionProgram->setUniform1i("vertexGraphColorEndOffset", vertexGraphColorEndOffset[graphColorIteration]);
			mSubdivisionProgram->compute1D(numThreads);

			glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
		}
	}

	mSubdivisionProgram->unbind();

	mNumReusableTriangleIndices = triangleNewFreeInterleaved;
	mUnusedTrianglesOffset = triangleNewFreeEnd;
	mNumReusableVertexIndices = vertexNewFreeInterleaved;
	mUnusedVerticesOffset = vertexNewFreeEnd;

	return true;
}

//=================================================================================================
/**
*	Merges all triangles flagged for merge.
*
*	@author	Daniel Cornel
*
*	@param[out]	outNumSelectedTriangles	The number of triangles selected for subdivision
*/
//=================================================================================================
bool TessellationCornelEtAl2022::mergeTriangles(const Camera& camera, int& outNumSelectedTriangles)
{
	for (int graphColorIteration = 0; graphColorIteration < 4; graphColorIteration++)
	{
		glClearNamedBufferData(mGraphColorCounters[graphColorIteration], GL_R32I, GL_RED_INTEGER, GL_INT, nullptr);
	}

	mMergeTriangleSelectionProgram->bind();
	mMergeTriangleSelectionProgram->setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	mMergeTriangleSelectionProgram->setSSBO("firstChildTriangleBuffer", mFirstChildTriangleBuffer);
	mMergeTriangleSelectionProgram->setSSBO("parentTriangleBuffer", mParentTriangleBuffer);
	mMergeTriangleSelectionProgram->setSSBO("adjacentTrianglesBuffer", mAdjacentTrianglesBuffer);
	mMergeTriangleSelectionProgram->setSSBO("triangleGraphColorBuffer", mTriangleGraphColorBuffer);
	mMergeTriangleSelectionProgram->setSSBO("graphColor0Buffer", mGraphColorBuffers[0]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor0Counter", mGraphColorCounters[0]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor1Buffer", mGraphColorBuffers[1]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor1Counter", mGraphColorCounters[1]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor2Buffer", mGraphColorBuffers[2]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor2Counter", mGraphColorCounters[2]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor3Buffer", mGraphColorBuffers[3]);
	mMergeTriangleSelectionProgram->setSSBO("graphColor3Counter", mGraphColorCounters[3]);
	mMergeTriangleSelectionProgram->compute1D(mUnusedTrianglesOffset);
	mMergeTriangleSelectionProgram->unbind();

	glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT | GL_BUFFER_UPDATE_BARRIER_BIT);

	Vector4i selectedTrianglesPerGraphColor = Vector4i::Zero();

	for (int graphColorIteration = 0; graphColorIteration < 4; graphColorIteration++)
	{
		glGetNamedBufferSubData(mGraphColorCounters[graphColorIteration], 0, sizeof(GLint), &selectedTrianglesPerGraphColor[graphColorIteration]);
	}

	int totalNumberOfTriangles = selectedTrianglesPerGraphColor.sum();

	outNumSelectedTriangles = totalNumberOfTriangles;

	if (totalNumberOfTriangles <= 0)
	{
		return false;
	}

	int numReusableTriangleIndices = mNumReusableTriangleIndices;
	int vertexNumInterleaved = mNumReusableVertexIndices;

	Vector4i freeTriangleIndicesOffsets = Vector4i::Zero();
	freeTriangleIndicesOffsets.x() = numReusableTriangleIndices;
	freeTriangleIndicesOffsets.y() = freeTriangleIndicesOffsets.x() + selectedTrianglesPerGraphColor.x();
	freeTriangleIndicesOffsets.z() = freeTriangleIndicesOffsets.y() + selectedTrianglesPerGraphColor.y();
	freeTriangleIndicesOffsets.w() = freeTriangleIndicesOffsets.z() + selectedTrianglesPerGraphColor.z();

	Vector4i freeVertexIndicesOffsets = Vector4i::Zero();
	freeVertexIndicesOffsets.x() = vertexNumInterleaved;
	freeVertexIndicesOffsets.y() = freeVertexIndicesOffsets.x() + selectedTrianglesPerGraphColor.x();
	freeVertexIndicesOffsets.z() = freeVertexIndicesOffsets.y() + selectedTrianglesPerGraphColor.y();
	freeVertexIndicesOffsets.w() = freeVertexIndicesOffsets.z() + selectedTrianglesPerGraphColor.z();

	mMergeProgram->bind();
	mMergeProgram->setCamera(camera);
	mMergeProgram->setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	mMergeProgram->setSSBO("firstChildTriangleBuffer", mFirstChildTriangleBuffer);
	mMergeProgram->setSSBO("parentTriangleBuffer", mParentTriangleBuffer);
	mMergeProgram->setSSBO("adjacentTrianglesBuffer", mAdjacentTrianglesBuffer);
	mMergeProgram->setSSBO("vertexIndicesBuffer", mVertexIndicesBuffer);
	mMergeProgram->setSSBO("freeTriangleIndicesBuffer", mFreeTriangleIndicesBuffer);
	mMergeProgram->setSSBO("freeVertexIndicesBuffer", mFreeVertexIndicesBuffer);

	bool synchronizationRequired = false;

	for (int graphColor = 0; graphColor < 8; graphColor += 2)
	{
		int graphColorIteration = graphColor >> 1;
		int numThreads = selectedTrianglesPerGraphColor[graphColorIteration];

		if (numThreads > 0)
		{
			mMergeProgram->setSSBO("graphColorBuffer", mGraphColorBuffers[graphColorIteration]);
			mMergeProgram->setUniform1i("freeTriangleIndicesOffset", freeTriangleIndicesOffsets[graphColorIteration]);
			mMergeProgram->setUniform1i("freeVertexIndicesOffset", freeVertexIndicesOffsets[graphColorIteration]);
			mMergeProgram->compute1D(numThreads);

			glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT | GL_SHADER_STORAGE_BARRIER_BIT);
		}
	}

	mMergeProgram->unbind();

	mNumReusableTriangleIndices = numReusableTriangleIndices + totalNumberOfTriangles;
	mNumReusableVertexIndices = vertexNumInterleaved + totalNumberOfTriangles;

	return true;
}

//=================================================================================================
/**
*	Compact all currently visible triangles for draw call.
*
*	@author	Daniel Cornel
*/
//=================================================================================================
void TessellationCornelEtAl2022::compactTriangles(const Camera& camera)
{
	const DrawElementsIndirectCommand drawCommand(0, 1, 0, 0, 0);

	glNamedBufferSubData(mIndirectDrawBuffer, 0, sizeof(drawCommand), &drawCommand);

	mCompactionProgram->bind();
	mCompactionProgram->setCamera(camera);
	mCompactionProgram->setSSBO("vertexBuffer", mVertexBuffer);
	mCompactionProgram->setSSBO("subdivisionFlagsBuffer", mSubdivisionFlagsBuffer);
	mCompactionProgram->setSSBO("firstChildTriangleBuffer", mFirstChildTriangleBuffer);
	mCompactionProgram->setSSBO("vertexIndicesBuffer", mVertexIndicesBuffer);
	mCompactionProgram->setSSBO("indirectDrawBuffer", mIndirectDrawBuffer);
	mCompactionProgram->setSSBO("visibleTriangleBuffer", mVisibleTriangleBuffer);
	mCompactionProgram->compute1D(mUnusedTrianglesOffset);
	mCompactionProgram->unbind();

	glMemoryBarrier(GL_ELEMENT_ARRAY_BARRIER_BIT | GL_SHADER_STORAGE_BARRIER_BIT | GL_COMMAND_BARRIER_BIT);
}