#include "TessellationKhouryEtAl2019.h"

#include <ppl.h>

#include <GL/glew.h>

#include "../Heightfield.h"
#include "ShaderCodesKhouryEtAl2019.h"

#include "../PipelineStatisticsQuery.h"

using namespace Eigen;

const float averageEdgeLengthInPixels = 10.0f;
const int bufferCapacity = 1u << 24; // Note: This value is also defined in the shader codes

//=================================================================================================
/**
*	Constructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationKhouryEtAl2019::TessellationKhouryEtAl2019(const Heightfield& heightfield) : AbstractTessellation(heightfield, "KhouryEtAl2019")
{
	Vector2f domainUpper = mHeightfield.mOrigin + mHeightfield.mCellSize.cwiseProduct(mHeightfield.mDimensions.cast<float>());

	const ChunkDescriptor& chunkDescriptor = mHeightfield.mChunkDescriptor;

	//=================================================================================================
	/**
	*	Create vertex buffer for the tessellation grid
	*/
	//=================================================================================================

	std::vector<Vector3f> gridVertices((chunkDescriptor.mChunkGridDimensions.x() + 1) * (chunkDescriptor.mChunkGridDimensions.y() + 1));

	Concurrency::parallel_for(0, static_cast<int>(gridVertices.size()), [&](int vertexIndex)
	{
		int x = vertexIndex % (chunkDescriptor.mChunkGridDimensions.x() + 1);
		int y = vertexIndex / (chunkDescriptor.mChunkGridDimensions.x() + 1);

		gridVertices[vertexIndex] = Vector3f(
			mHeightfield.mOrigin.x() + static_cast<float>(x) * chunkDescriptor.mChunkCellSize.x(),
			mHeightfield.mOrigin.y() + static_cast<float>(y) * chunkDescriptor.mChunkCellSize.y(),
			0.0f);
	});

	glCreateBuffers(1, &mGridVertexBuffer);
	glNamedBufferData(mGridVertexBuffer, sizeof(Vector3f) * gridVertices.size(), gridVertices.data(), GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create index buffer for the tessellation grid
	*/
	//=================================================================================================

	std::vector<GLuint> gridIndices(6 * chunkDescriptor.mChunkGridDimensions.x() * chunkDescriptor.mChunkGridDimensions.y());

	Concurrency::parallel_for(0, chunkDescriptor.mChunkGridDimensions.x() * chunkDescriptor.mChunkGridDimensions.y(), [&](int cellIndex)
	{
		int x = cellIndex % chunkDescriptor.mChunkGridDimensions.x();
		int y = cellIndex / chunkDescriptor.mChunkGridDimensions.x();

		int bottomLeftVertexIndex = y * (chunkDescriptor.mChunkGridDimensions.x() + 1) + x;
		int bottomRightVertexIndex = bottomLeftVertexIndex + 1;
		int topLeftVertexIndex = (y + 1) * (chunkDescriptor.mChunkGridDimensions.x() + 1) + x;
		int topRightVertexIndex = topLeftVertexIndex + 1;

		gridIndices[6 * cellIndex] = bottomLeftVertexIndex;
		gridIndices[6 * cellIndex + 1] = bottomRightVertexIndex;
		gridIndices[6 * cellIndex + 2] = topLeftVertexIndex;
		gridIndices[6 * cellIndex + 3] = topRightVertexIndex;
		gridIndices[6 * cellIndex + 4] = topLeftVertexIndex;
		gridIndices[6 * cellIndex + 5] = bottomRightVertexIndex;
	});

	GLuint numInitialTriangles = static_cast<GLuint>(gridIndices.size() / 3);

	glCreateBuffers(1, &mGridIndexBuffer);
	glNamedBufferData(mGridIndexBuffer, sizeof(GLuint) * gridIndices.size(), gridIndices.data(), GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create vertex buffer for the triangle instance
	*/
	//=================================================================================================

	std::vector<Vector2f> instanceVertices
	{
		{ 0.0f, 0.0f },
		{ 1.0f, 0.0f },
		{ 0.0f, 1.0f }
	};

	glCreateBuffers(1, &mTriangleInstanceVertexBuffer);
	glNamedBufferData(mTriangleInstanceVertexBuffer, sizeof(Vector2f) * instanceVertices.size(), instanceVertices.data(), GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create index buffer for the triangle instance
	*/
	//=================================================================================================

	Vector3i instanceIndices(0, 1, 2);

	glCreateBuffers(1, &mTriangleInstanceIndexBuffer);
	glNamedBufferData(mTriangleInstanceIndexBuffer, sizeof(Vector3i), instanceIndices.data(), GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create subdivision buffers
	*/
	//=================================================================================================

	std::vector<Vector2ui> subdivisionKeys(numInitialTriangles);

	for (GLuint triangleIndex = 0u; triangleIndex < numInitialTriangles; triangleIndex++)
	{
		subdivisionKeys[triangleIndex] = Vector2ui(triangleIndex, 1u);
	}

	glCreateBuffers(1, &mSubdivisionBuffer1);
	glNamedBufferData(mSubdivisionBuffer1, sizeof(Vector2ui) * bufferCapacity, nullptr, GL_DYNAMIC_DRAW);
	glNamedBufferSubData(mSubdivisionBuffer1, 0, sizeof(Vector2ui) * subdivisionKeys.size(), subdivisionKeys.data());

	glCreateBuffers(1, &mSubdivisionBuffer2);
	glNamedBufferData(mSubdivisionBuffer2, sizeof(Vector2ui) * bufferCapacity, nullptr, GL_DYNAMIC_DRAW);
	glNamedBufferSubData(mSubdivisionBuffer2, 0, sizeof(Vector2ui) * subdivisionKeys.size(), subdivisionKeys.data());

	glCreateBuffers(1, &mCulledTrianglesBuffer);
	glNamedBufferData(mCulledTrianglesBuffer, sizeof(Vector2ui) * bufferCapacity, nullptr, GL_DYNAMIC_DRAW);

	//=================================================================================================
	/**
	*	Create the programs
	*/
	//=================================================================================================

	mLoDProgram = std::make_unique<Program>(ShaderCodesKhouryEtAl2019::LoDCompute);
	mLoDProgram->setUniform1f("heightfieldScaleFactor", mHeightfield.mScaleFactor);
	mLoDProgram->setUniform2i("gridDimensions", mHeightfield.mDimensions);
	mLoDProgram->setUniform2f("domainOrigin", mHeightfield.mOrigin);
	mLoDProgram->setUniform2f("cellSize", mHeightfield.mCellSize);

	mIndirectProgram = std::make_unique<Program>(ShaderCodesKhouryEtAl2019::IndirectCompute);

	mTerrainProgram = std::make_unique<Program>(
		ShaderCodesKhouryEtAl2019::TerrainVertex,
		ShaderCodesKhouryEtAl2019::CommonGeometryShader,
		ShaderCodesKhouryEtAl2019::CommonFragmentShader);
	mTerrainProgram->setUniform1f("heightfieldScaleFactor", mHeightfield.mScaleFactor);
	mTerrainProgram->setUniform2i("gridDimensions", mHeightfield.mDimensions);
	mTerrainProgram->setUniform2f("domainOrigin", mHeightfield.mOrigin);
	mTerrainProgram->setUniform2f("cellSize", mHeightfield.mCellSize);

	//=================================================================================================
	/**
	*	Create the vertex array object
	*/
	//=================================================================================================

	glCreateVertexArrays(1, &mVAO);
	glVertexArrayElementBuffer(mVAO, mTriangleInstanceIndexBuffer);

	GLint positionAttributeLocation = mTerrainProgram->getAttributeLocation("inPosition");
	glEnableVertexArrayAttrib(mVAO, positionAttributeLocation);
	glVertexArrayAttribFormat(mVAO, positionAttributeLocation, 2, GL_FLOAT, GL_FALSE, 0);
	glVertexArrayVertexBuffer(mVAO, positionAttributeLocation, mTriangleInstanceVertexBuffer, 0, sizeof(Vector2f));

	//=================================================================================================
	/**
	*	Create the indirect draw resources
	*/
	//=================================================================================================

	// This struct is defined by OpenGL
	struct DispatchIndirectCommand
	{
		GLuint num_groups_x = 1u;
		GLuint num_groups_y = 1u;
		GLuint num_groups_z = 1u;
	} dispatchIndirectCommand;
	dispatchIndirectCommand.num_groups_x = numInitialTriangles / 32 + 1u;

	glCreateBuffers(1, &mIndirectDispatchBuffer);
	glNamedBufferData(mIndirectDispatchBuffer, sizeof(DispatchIndirectCommand), &dispatchIndirectCommand, GL_DYNAMIC_DRAW);

	// This struct is defined by OpenGL
	struct DrawElementsIndirectCommand
	{
		GLuint count = 3u;
		GLuint primCount = 0u;
		GLuint firstIndex = 0u;
		GLuint baseVertex = 0u;
		GLuint baseInstance = 0u;
	} drawElementsIndirectCommand;

	glCreateBuffers(1, &mIndirectDrawBuffer);
	glNamedBufferData(mIndirectDrawBuffer, sizeof(DrawElementsIndirectCommand), &drawElementsIndirectCommand, GL_DYNAMIC_DRAW);

	glCreateBuffers(1, &mNumInputTrianglesBuffer);
	glNamedBufferData(mNumInputTrianglesBuffer, sizeof(GLuint), &numInitialTriangles, GL_DYNAMIC_DRAW);

	glCreateBuffers(1, &mNumOutputTrianglesBuffer);
	glNamedBufferData(mNumOutputTrianglesBuffer, sizeof(GLuint), nullptr, GL_DYNAMIC_DRAW);
	glClearNamedBufferData(mNumOutputTrianglesBuffer, GL_R32UI, GL_RED_INTEGER, GL_UNSIGNED_INT, nullptr);

	glCreateBuffers(1, &mChangesAtomicCounterBuffer);
	glNamedBufferData(mChangesAtomicCounterBuffer, sizeof(GLint), nullptr, GL_DYNAMIC_DRAW);
}

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationKhouryEtAl2019::~TessellationKhouryEtAl2019()
{
	glDeleteBuffers(1, &mTriangleInstanceVertexBuffer);
	glDeleteBuffers(1, &mTriangleInstanceIndexBuffer);
	glDeleteBuffers(1, &mGridVertexBuffer);
	glDeleteBuffers(1, &mGridIndexBuffer);
	glDeleteBuffers(1, &mSubdivisionBuffer1);
	glDeleteBuffers(1, &mSubdivisionBuffer2);
	glDeleteBuffers(1, &mCulledTrianglesBuffer);
	glDeleteBuffers(1, &mNumInputTrianglesBuffer);
	glDeleteBuffers(1, &mNumOutputTrianglesBuffer);
	glDeleteBuffers(1, &mIndirectDispatchBuffer);
	glDeleteBuffers(1, &mIndirectDrawBuffer);
	glDeleteBuffers(1, &mChangesAtomicCounterBuffer);

	glDeleteVertexArrays(1, &mVAO);
}

//=================================================================================================
/**
*	Performs a per-frame update of the tessellation.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationKhouryEtAl2019::update(const Camera& camera)
{
	float lodFactor = averageEdgeLengthInPixels * 2.0f * std::tan(0.5f * camera.mVerticalFieldOfView) /
		camera.mViewPortSize.x() / mHeightfield.mChunkDescriptor.mChunkCellSize.x();

	mLoDProgram->setUniform1f("lodFactor", lodFactor);
	mLoDProgram->setCamera(camera);

	bool iterate = false;

	do
	{
		glClearNamedBufferData(mChangesAtomicCounterBuffer, GL_R32I, GL_RED_INTEGER, GL_INT, nullptr);

		subdivideMerge(camera);

		GLint numChanges = 0;
		glGetNamedBufferSubData(mChangesAtomicCounterBuffer, 0, sizeof(GLint), &numChanges);

		iterate = (numChanges >= 100);

		if (iterate)
		{
			resetBuffers();

			glGetNamedBufferSubData(mChangesAtomicCounterBuffer, 0, sizeof(GLint), &numChanges);
		}

	} while (iterate);

}

//=================================================================================================
/**
*	Renders the geometry.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationKhouryEtAl2019::render(const Camera& camera)
{
	mTerrainProgram->bind();
	mTerrainProgram->setCamera(camera);
	mTerrainProgram->setSSBO("vertexBuffer", mGridVertexBuffer);
	mTerrainProgram->setSSBO("indexBuffer", mGridIndexBuffer);
	mTerrainProgram->setSSBO("outputSubdivisionBuffer", (mPingPong) ? mSubdivisionBuffer2 : mSubdivisionBuffer1);
	mTerrainProgram->setSSBO("culledTrianglesBuffer", mCulledTrianglesBuffer);
	mTerrainProgram->setSSBO("numOutputTrianglesBuffer", mNumOutputTrianglesBuffer);
	mTerrainProgram->setSSBO("heightfieldBuffer", mHeightfield.mBuffer);

	glBindVertexArray(mVAO);

	glBindBuffer(GL_DRAW_INDIRECT_BUFFER, mIndirectDrawBuffer);
	glDrawElementsIndirect(GL_TRIANGLES, GL_UNSIGNED_INT, 0);
	glMemoryBarrier(GL_ALL_BARRIER_BITS);

	glBindVertexArray(0u);

	mTerrainProgram->unbind();

	resetBuffers();
}

//=================================================================================================
/**
*	Performs one subdivision/merge iteration.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationKhouryEtAl2019::subdivideMerge(const Camera& camera)
{
	mLoDProgram->bind();
	mLoDProgram->setSSBO("vertexBuffer", mGridVertexBuffer);
	mLoDProgram->setSSBO("indexBuffer", mGridIndexBuffer);
	mLoDProgram->setSSBO("inputSubdivisionBuffer", (mPingPong) ? mSubdivisionBuffer1 : mSubdivisionBuffer2);
	mLoDProgram->setSSBO("outputSubdivisionBuffer", (mPingPong) ? mSubdivisionBuffer2 : mSubdivisionBuffer1);
	mLoDProgram->setSSBO("culledTrianglesBuffer", mCulledTrianglesBuffer);
	mLoDProgram->setSSBO("indirectDispatchBuffer", mIndirectDispatchBuffer);
	mLoDProgram->setSSBO("indirectDrawBuffer", mIndirectDrawBuffer);
	mLoDProgram->setSSBO("numInputTrianglesBuffer", mNumInputTrianglesBuffer);
	mLoDProgram->setSSBO("numOutputTrianglesBuffer", mNumOutputTrianglesBuffer);
	mLoDProgram->setSSBO("changesAtomicCounterBuffer", mChangesAtomicCounterBuffer);
	mLoDProgram->setSSBO("heightfieldBuffer", mHeightfield.mBuffer);

	glBindBuffer(GL_DISPATCH_INDIRECT_BUFFER, mIndirectDispatchBuffer);
	glDispatchComputeIndirect(0);
	glMemoryBarrier(GL_ALL_BARRIER_BITS);

	mLoDProgram->unbind();
}

//=================================================================================================
/**
*	Resets the temporary buffers and creates the command for the next update dispatch.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationKhouryEtAl2019::resetBuffers()
{
	mIndirectProgram->bind();
	mIndirectProgram->setSSBO("indirectDispatchBuffer", mIndirectDispatchBuffer);
	mIndirectProgram->setSSBO("indirectDrawBuffer", mIndirectDrawBuffer);
	mIndirectProgram->setSSBO("numInputTrianglesBuffer", mNumInputTrianglesBuffer);
	mIndirectProgram->setSSBO("numOutputTrianglesBuffer", mNumOutputTrianglesBuffer);

	glDispatchCompute(1, 1, 1);
	glMemoryBarrier(GL_ALL_BARRIER_BITS);

	mIndirectProgram->unbind();

	mPingPong = !mPingPong;
}