#include "Program.h"

#include <iostream>
#include <sstream>
#include <vector>

#include "Camera.h"

//=================================================================================================
/**
*	Constructor for a rendering program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Program::Program(
	const std::vector<const char*>& vertexShaderCode,
	const std::vector<const char*>& fragmentShaderCode)
{
	mVertexShaderHandle = createShader(GL_VERTEX_SHADER, makeString(vertexShaderCode).c_str());
	mFragmentShaderHandle = createShader(GL_FRAGMENT_SHADER, makeString(fragmentShaderCode).c_str());

	mHandle = glCreateProgram();

	glAttachShader(mHandle, mVertexShaderHandle);
	glAttachShader(mHandle, mFragmentShaderHandle);

	glBindFragDataLocation(mHandle, 0, "outColor");

	link();
}

//=================================================================================================
/**
*	Constructor for a rendering program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Program::Program(
	const std::vector<const char*>& vertexShaderCode,
	const std::vector<const char*>& geometryShaderCode,
	const std::vector<const char*>& fragmentShaderCode)
{
	mVertexShaderHandle = createShader(GL_VERTEX_SHADER, makeString(vertexShaderCode).c_str());
	mGeometryShaderHandle = createShader(GL_GEOMETRY_SHADER, makeString(geometryShaderCode).c_str());
	mFragmentShaderHandle = createShader(GL_FRAGMENT_SHADER, makeString(fragmentShaderCode).c_str());

	mHandle = glCreateProgram();

	glAttachShader(mHandle, mVertexShaderHandle);
	glAttachShader(mHandle, mGeometryShaderHandle);
	glAttachShader(mHandle, mFragmentShaderHandle);

	glBindFragDataLocation(mHandle, 0, "outColor");

	link();
}

//=================================================================================================
/**
*	Constructor for a tessellation program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Program::Program(
	const std::vector<const char*>& vertexShaderCode,
	const std::vector<const char*>& controlShaderCode,
	const std::vector<const char*>& evaluationShaderCode,
	const std::vector<const char*>& geometryShaderCode,
	const std::vector<const char*>& fragmentShaderCode)
{
	mVertexShaderHandle = createShader(GL_VERTEX_SHADER, makeString(vertexShaderCode).c_str());
	mControlShaderHandle = createShader(GL_TESS_CONTROL_SHADER, makeString(controlShaderCode).c_str());
	mEvaluationShaderHandle = createShader(GL_TESS_EVALUATION_SHADER, makeString(evaluationShaderCode).c_str());
	mGeometryShaderHandle = createShader(GL_GEOMETRY_SHADER, makeString(geometryShaderCode).c_str());
	mFragmentShaderHandle = createShader(GL_FRAGMENT_SHADER, makeString(fragmentShaderCode).c_str());

	mHandle = glCreateProgram();

	glAttachShader(mHandle, mVertexShaderHandle);
	glAttachShader(mHandle, mControlShaderHandle);
	glAttachShader(mHandle, mEvaluationShaderHandle);
	glAttachShader(mHandle, mGeometryShaderHandle);
	glAttachShader(mHandle, mFragmentShaderHandle);

	glBindFragDataLocation(mHandle, 0, "outColor");

	link();
}

//=================================================================================================
/**
*	Constructor for a compute program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Program::Program(const std::vector<const char*>& computeShaderCode)
{
	mComputeShaderHandle = createShader(GL_COMPUTE_SHADER, makeString(computeShaderCode).c_str());

	mHandle = glCreateProgram();

	glAttachShader(mHandle, mComputeShaderHandle);

	link();
}

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Program::~Program()
{
	unbind();

	glDeleteShader(mVertexShaderHandle);
	glDeleteShader(mControlShaderHandle);
	glDeleteShader(mEvaluationShaderHandle);
	glDeleteShader(mGeometryShaderHandle);
	glDeleteShader(mFragmentShaderHandle);
	glDeleteShader(mComputeShaderHandle);

	glDeleteProgram(mHandle);
}

//====================================================================================================
/**
*	Concatenates the individual shader pieces to one string. Shader codes might be given as vector of
*	strings rather than a single one to easily handle shared use of code without GLSL includes.
*
*	@author	Daniel Cornel
*/
//====================================================================================================
std::string Program::makeString(const std::vector<const char*> shaderCode)
{
	std::stringstream sstr;

	for (const char* part : shaderCode)
	{
		sstr << part << "\n";
	}

	return sstr.str();
}

//====================================================================================================
/**
*	Creates and compiles a shader object from the provided source code.
*
*	@author	Daniel Cornel
*/
//====================================================================================================
GLuint Program::createShader(GLenum shaderType, const char* shaderCode) const
{
	GLuint shaderHandle = glCreateShader(shaderType);

	glShaderSource(shaderHandle, 1, &shaderCode, NULL);
	glCompileShader(shaderHandle);

	GLint shaderCompileStatus = GL_FALSE;
	glGetShaderiv(shaderHandle, GL_COMPILE_STATUS, &shaderCompileStatus);

	int infologLength = 0;
	glGetShaderiv(shaderHandle, GL_INFO_LOG_LENGTH, &infologLength);

	if ((shaderCompileStatus != GL_TRUE) || (infologLength > 0))
	{
		std::stringstream error;
		error << "Shader compilation failed:" << std::endl;

		if (infologLength > 0)
		{
			std::vector<char> infoLog(infologLength + 1, 0);
			int charsWritten = 0;

			glGetShaderInfoLog(shaderHandle, infologLength, &charsWritten, &infoLog[0]);

			std::string infoLogString(&infoLog[0]);

			error << infoLogString;
			error << std::endl;
		}

		std::string errorString = error.str();

		throw std::runtime_error(errorString);
	}

	return shaderHandle;
}

//=================================================================================================
/**
*	Links the given shaders into the shader program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::link()
{
	glProgramParameteri(mHandle, GL_PROGRAM_BINARY_RETRIEVABLE_HINT, GL_TRUE);

	glLinkProgram(mHandle);

	// Throw exception if linking failed
	GLint status = GL_FALSE;
	glGetProgramiv(mHandle, GL_LINK_STATUS, &status);

	int infologLength = 0;
	glGetProgramiv(mHandle, GL_INFO_LOG_LENGTH, &infologLength);

	if ((status != GL_TRUE) || (infologLength > 0))
	{
		std::stringstream error;
		error << "Shader program linking failed:" << std::endl;

		if (infologLength > 0)
		{
			std::vector<char> infoLog(infologLength + 1, 0);
			int charsWritten = 0;

			glGetProgramInfoLog(mHandle, infologLength, &charsWritten, &infoLog[0]);
			std::string infoLogString(&infoLog[0]);

			error << std::endl;
			error << infoLogString;
		}

		std::string errorString = error.str();

		throw std::runtime_error(errorString);
	}

	getIntrospectionData();
}

//=================================================================================================
/**
*	Acquires and updates data about the linked program using the introspection API.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::getIntrospectionData()
{
	{
		mUniforms.clear();

		GLint numUniforms = 0;
		glGetProgramInterfaceiv(mHandle, GL_UNIFORM, GL_ACTIVE_RESOURCES, &numUniforms);

		std::vector<GLenum> properties = { GL_TYPE, GL_NAME_LENGTH, GL_LOCATION, GL_ARRAY_SIZE };

		for (int iUniform = 0; iUniform < numUniforms; iUniform++)
		{
			std::vector<GLint> values(4);
			glGetProgramResourceiv(mHandle, GL_UNIFORM, iUniform, static_cast<GLsizei>(properties.size()), properties.data(), static_cast<GLsizei>(values.size()), nullptr, values.data());

			const GLint& type = values[0];
			GLint& nameLength = values[1];
			const GLint& location = values[2];
			const GLint& arraySize = values[3];

			if ((location == -1) || (nameLength <= 0))
			{
				continue;
			}

			std::string name(nameLength, '\0');
			glGetProgramResourceName(mHandle, GL_UNIFORM, iUniform, GLsizei(name.size()), &nameLength, &name[0]);

			name.erase(name.begin() + nameLength, name.end());

			auto found = name.find_first_of('[');

			if ((found != std::string::npos) && (found < name.length()))
			{
				name = name.substr(0, found);
			}

			mUniforms[name] = IntrospectionUniform(location, type, (arraySize > 1), arraySize, name);
		}
	}

	{
		mInputs.clear();

		GLint numInputs = 0;
		glGetProgramInterfaceiv(mHandle, GL_PROGRAM_INPUT, GL_ACTIVE_RESOURCES, &numInputs);

		std::vector<GLenum> properties = { GL_TYPE, GL_NAME_LENGTH, GL_LOCATION };

		for (int iInput = 0; iInput < numInputs; iInput++)
		{
			std::vector<GLint> values(3);
			glGetProgramResourceiv(mHandle, GL_PROGRAM_INPUT, iInput, static_cast<GLsizei>(properties.size()), properties.data(), static_cast<GLsizei>(values.size()), nullptr, values.data());

			const GLint& type = values[0];
			GLint& nameLength = values[1];
			const GLint& location = values[2];

			if ((location == -1) || (nameLength <= 0))
			{
				continue;
			}

			std::string name(nameLength, '\0');
			glGetProgramResourceName(mHandle, GL_PROGRAM_INPUT, iInput, GLsizei(name.length()), &nameLength, &name[0]);
			name.erase(name.begin() + nameLength, name.end());

			mInputs[name] = IntrospectionInput(location, type, name);
		}
	}

	{
		mInterfaceBlocks.clear();

		GLint numInterfaceBlocks = 0;
		glGetProgramInterfaceiv(mHandle, GL_SHADER_STORAGE_BLOCK, GL_ACTIVE_RESOURCES, &numInterfaceBlocks);

		const GLenum blockNameProperty = GL_NAME_LENGTH;

		for (int iInterfaceBlock = 0; iInterfaceBlock < numInterfaceBlocks; iInterfaceBlock++)
		{
			GLint nameLength = 0;
			glGetProgramResourceiv(mHandle, GL_SHADER_STORAGE_BLOCK, iInterfaceBlock, 1, &blockNameProperty, 1, nullptr, &nameLength);

			std::string name(nameLength, '\0');
			glGetProgramResourceName(mHandle, GL_SHADER_STORAGE_BLOCK, iInterfaceBlock, GLsizei(name.size()), &nameLength, &name[0]);

			GLint location = glGetProgramResourceIndex(mHandle, GL_SHADER_STORAGE_BLOCK, name.c_str());

			if (location == GL_INVALID_INDEX)
			{
				location = -1;
			}

			if ((location == -1) || (nameLength <= 0))
			{
				continue;
			}

			name.erase(name.begin() + nameLength, name.end());

			mInterfaceBlocks[name] = IntrospectionInterfaceBlock(location, name);
		}
	}
}

//=================================================================================================
/**
*	Binds this shader program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::bind()
{
	glUseProgram(mHandle);
}

//=================================================================================================
/**
*	Unbinds this shader program.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::unbind()
{
	glUseProgram(0u);
}

//=================================================================================================
/**
*	Returns the location of the attribute with the provided name in the shader.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
GLint Program::getAttributeLocation(const std::string& name) const
{
	auto itr = mInputs.find(name);

	if (itr == mInputs.end())
	{
		return -1;
	}
	else
	{
		return itr->second.mLocation;
	}
}

//=================================================================================================
/**
*	Returns the location to the uniform with the provided name in the shader.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
GLint Program::getUniformLocation(const std::string& name) const
{
	auto itr = mUniforms.find(name);

	if (itr == mUniforms.end())
	{
		return -1;
	}
	else
	{
		return itr->second.mLocation;
	}
}

//=================================================================================================
/**
*	Returns the index to the interface block with the provided name in the shader.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
GLint Program::getInterfaceBlockIndex(const std::string& name) const
{
	auto itr = mInterfaceBlocks.find(name);

	if (itr == mInterfaceBlocks.end())
	{
		return -1;
	}
	else
	{
		return itr->second.mLocation;
	}
}

//=================================================================================================
/**
*	Retrieves the block binding index for the provided interface block.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
GLuint Program::getBlockBinding(GLuint interfaceBlockIndex)
{
	auto it = mBlockBindings.find(interfaceBlockIndex);

	if (it == mBlockBindings.end())
	{
		mBlockBindings[interfaceBlockIndex] = mBlockBindingCounter++;

		it = mBlockBindings.find(interfaceBlockIndex);

		glShaderStorageBlockBinding(mHandle, interfaceBlockIndex, it->second);
	}

	return it->second;
}

//=================================================================================================
/**
*	Binds the shader storage buffer to the interface block index with the provided name.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setSSBO(const std::string& name, GLuint bufferHandle)
{
	std::string interfaceBlockName = name + "Block";

	GLint interfaceBlockIndex = getInterfaceBlockIndex(interfaceBlockName);

	if (interfaceBlockIndex >= 0)
	{
		GLuint blockBindingIndex = getBlockBinding(interfaceBlockIndex);

		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, blockBindingIndex, bufferHandle);
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniformBool(const std::string& name, bool x)
{
	setUniform1i(name, x ? 1 : 0);
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniform1f(const std::string& name, float x)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniform1f(mHandle, uniformLocation, x);
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniform1i(const std::string& name, int x)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniform1i(mHandle, uniformLocation, x);
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniform2f(const std::string& name, const Eigen::Vector2f& x)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniform2fv(mHandle, uniformLocation, 1, x.data());
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniform2fArray(const std::string& name, const std::vector<Eigen::Vector2f>& x)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniform2fv(mHandle, uniformLocation, GLsizei(x.size()), x.data()->data());
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniform2i(const std::string& name, const Eigen::Vector2i& x)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniform2iv(mHandle, uniformLocation, 1, x.data());
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setUniform4f(const std::string& name, const Eigen::Vector4f& x)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniform4fv(mHandle, uniformLocation, 1, x.data());
	}
}

//=================================================================================================
/**
*	Sets the uniform to the provided value.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setMatrix4f(const std::string& name, const Eigen::Matrix4f& m)
{
	GLint uniformLocation = getUniformLocation(name);

	if (uniformLocation >= 0)
	{
		glProgramUniformMatrix4fv(mHandle, uniformLocation, 1, GL_FALSE, m.data());
	}
}

//=================================================================================================
/**
*	Sets all camera uniforms.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::setCamera(const Camera& camera)
{
	setMatrix4f("viewMatrix", camera.mViewMatrix);
	setMatrix4f("projectionMatrix", camera.mProjectionMatrix);
	setMatrix4f("viewProjectionMatrix", camera.mViewProjectionMatrix);
	setMatrix4f("inverseViewMatrix", camera.mInverseViewMatrix);
	setUniform2f("viewPortSize", camera.mViewPortSize);
	setUniform4f("cameraPosition", camera.mPosition);

	// http://fgiesen.wordpress.com/2012/08/31/frustum-planes-from-the-projection-matrix/
	std::vector<Eigen::Vector4f> frustumPlanes
	{
		camera.mViewProjectionMatrix.row(3) + camera.mViewProjectionMatrix.row(0), // Left
		camera.mViewProjectionMatrix.row(3) - camera.mViewProjectionMatrix.row(0), // Right
		camera.mViewProjectionMatrix.row(3) + camera.mViewProjectionMatrix.row(1), // Bottom
		camera.mViewProjectionMatrix.row(3) - camera.mViewProjectionMatrix.row(1), // Top
		camera.mViewProjectionMatrix.row(3) + camera.mViewProjectionMatrix.row(2), // Far
		camera.mViewProjectionMatrix.row(3) - camera.mViewProjectionMatrix.row(2) // Near
	};

	for (int i = 0; i < 6; i++)
	{
		Eigen::Vector4f& plane = frustumPlanes.at(i);
		plane /= plane.head<3>().norm();
	}

	setUniform4f("frustumPlaneLeft", frustumPlanes[0]);
	setUniform4f("frustumPlaneRight", frustumPlanes[1]);
	setUniform4f("frustumPlaneBottom", frustumPlanes[2]);
	setUniform4f("frustumPlaneTop", frustumPlanes[3]);
	setUniform4f("frustumPlaneFar", frustumPlanes[4]);
	setUniform4f("frustumPlaneNear", frustumPlanes[5]);
}

//=================================================================================================
/**
*	Assigns threads to workgroups and executes the compute shader.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void Program::compute1D(int numThreads)
{
	setUniform1i("numThreads", numThreads);

	Eigen::Vector3i workGroups(static_cast<GLint>(std::ceil(static_cast<double>(numThreads) / 32.0)), 1, 1);

	glDispatchCompute(workGroups.x(), workGroups.y(), workGroups.z());
}